const React = require("react");
const { PropTypes, Component } = React;
const { ViewPropTypes } = (ReactNative = require("react-native"));
const { Dimensions, FlatList, View, Animated, ScrollView, StyleSheet, InteractionManager, Platform } = ReactNative;
const TimerMixin = require("react-timer-mixin");

const SceneComponent = require("./SceneComponent");
const DefaultTabBar = require("./DefaultTabBar");
const ScrollableTabBar = require("./ScrollableTabBar");
const createReactClass = require("create-react-class");

const AnimatedFlatList = Animated.createAnimatedComponent(FlatList);

const ScrollableTabView = createReactClass({
  mixins: [TimerMixin],
  statics: {
    DefaultTabBar,
    ScrollableTabBar
  },

  propTypes: {
    // tabBarPosition: PropTypes.oneOf(['top', 'bottom', 'overlayTop', 'overlayBottom', ]),
    // initialPage: PropTypes.number,
    // page: PropTypes.number,
    // onChangeTab: PropTypes.func,
    // onScroll: PropTypes.func,
    // renderTabBar: PropTypes.any,
    // style: ViewPropTypes.style,
    // contentProps: PropTypes.object,
    // scrollWithoutAnimation: PropTypes.bool,
    // locked: PropTypes.bool,
    // prerenderingSiblingsNumber: PropTypes.number,
    // collapsableBar: PropTypes.node,
  },

  getDefaultProps() {
    return {
      tabBarPosition: "top",
      initialPage: 0,
      page: -1,
      onChangeTab: () => {},
      onScroll: () => {},
      contentProps: {},
      scrollWithoutAnimation: false,
      locked: false,
      prerenderingSiblingsNumber: 0
    };
  },

  getInitialState() {
    const width = Dimensions.get("window").width;
    return {
      currentPage: this.props.initialPage,
      scrollX: new Animated.Value(this.props.initialPage * width),
      scrollValue: new Animated.Value(this.props.initialPage),
      containerWidth: width,
      sceneKeys: this.newSceneKeys({ currentPage: this.props.initialPage })
    };
  },

  componentDidMount() {
    this.setTimeout(() => {
      InteractionManager.runAfterInteractions(() => {
        if (Platform.OS === "android") {
          this.goToPage(this.props.initialPage, false);
        }
      });
    }, 0);

    this.state.scrollX.addListener(({ value }) => {
      const scrollValue = value / this.state.containerWidth;
      this.state.scrollValue.setValue(scrollValue);
      this.props.onScroll(scrollValue);
    });
  },

  componentWillReceiveProps(props) {
    if (props.children !== this.props.children) {
      this.updateSceneKeys({ page: this.state.currentPage, children: props.children });
    }

    if (props.page >= 0 && props.page !== this.state.currentPage) {
      this.goToPage(props.page);
    }
  },

  goToPage(pageNumber, animated = !this.props.scrollWithoutAnimation) {
    const offset = pageNumber * this.state.containerWidth;
    if (this.scrollView && this.scrollView._component && this.scrollView._component.scrollTo) {
      this.scrollView._component.scrollTo({ x: offset, y: 0, animated });
    }

    const currentPage = this.state.currentPage;
    this.updateSceneKeys({
      page: pageNumber,
      callback: this._onChangeTab.bind(this, currentPage, pageNumber)
    });
  },

  renderTabBar(props) {
    if (this.props.renderTabBar === false) {
      return null;
    } else if (this.props.renderTabBar) {
      return React.cloneElement(this.props.renderTabBar(props), {...props, key: "tabbar"});
    } else {
      return <DefaultTabBar {...props} />;
    }
  },

  updateSceneKeys({ page, children = this.props.children, callback = () => {} }) {
    let newKeys = this.newSceneKeys({ previousKeys: this.state.sceneKeys, currentPage: page, children });
    this.setState({ currentPage: page, sceneKeys: newKeys }, callback);
  },

  newSceneKeys({ previousKeys = [], currentPage = 0, children = this.props.children }) {
    let newKeys = [];
    this._children(children).forEach((child, idx) => {
      let key = this._makeSceneKey(child, idx);
      if (this._keyExists(previousKeys, key) || this._shouldRenderSceneKey(idx, currentPage)) {
        newKeys.push(key);
      }
    });
    return newKeys;
  },

  _shouldRenderSceneKey(idx, currentPageKey) {
    let numOfSibling = this.props.prerenderingSiblingsNumber;
    return idx < currentPageKey + numOfSibling + 1 && idx > currentPageKey - numOfSibling - 1;
  },

  _keyExists(sceneKeys, key) {
    return sceneKeys.find(sceneKey => key === sceneKey);
  },

  _makeSceneKey(child, idx) {
    return child.props.tabLabel + "_" + idx;
  },

  renderScrollableContent() {
    const scenes = this._composeScenes();
    return (
      <Animated.ScrollView
        key="content"
        horizontal
        pagingEnabled
        automaticallyAdjustContentInsets={false}
        contentOffset={{ x: this.props.initialPage * this.state.containerWidth }}
        ref={scrollView => {
          this.scrollView = scrollView;
        }}
        onScroll={Animated.event(
          [
            {
              nativeEvent: { contentOffset: { x: this.state.scrollX } }
            }
          ],
          {
            useNativeDriver: true
          }
        )}
        onMomentumScrollBegin={this._onMomentumScrollBeginAndEnd}
        onMomentumScrollEnd={this._onMomentumScrollBeginAndEnd}
        scrollEventThrottle={16}
        scrollsToTop={false}
        showsHorizontalScrollIndicator={false}
        scrollEnabled={!this.props.locked}
        directionalLockEnabled
        alwaysBounceVertical={false}
        keyboardDismissMode="on-drag"
        {...this.props.contentProps}
      >
        {scenes}
      </Animated.ScrollView>
    );
  },

  _composeScenes() {
    return this._children().map((child, idx) => {
      let key = this._makeSceneKey(child, idx);
      let element;

      if (!!this.props.collapsibleElement) {
        element = this.state.currentPage === idx ? child : null;
      } else {
        element = this._keyExists(this.state.sceneKeys, key) ? child : <View tabLabel={child.props.tabLabel} />;
      }

      return (
        <SceneComponent
          key={child.key}
          shouldUpdated={!!this.props.collapsibleElement || this._shouldRenderSceneKey(idx, this.state.currentPage)}
          style={{ width: this.state.containerWidth }}
        >
          {element}
        </SceneComponent>
      );
    });
  },

  _onMomentumScrollBeginAndEnd(e) {
    const offsetX = e.nativeEvent.contentOffset.x;
    const page = Math.round(offsetX / this.state.containerWidth);
    if (this.state.currentPage !== page) {
      this._updateSelectedPage(page);
    }
  },

  _updateSelectedPage(nextPage) {
    let localNextPage = nextPage;
    if (typeof localNextPage === "object") {
      localNextPage = nextPage.nativeEvent.position;
    }

    const currentPage = this.state.currentPage;
    this.updateSceneKeys({
      page: localNextPage,
      callback: this._onChangeTab.bind(this, currentPage, localNextPage)
    });
  },

  _onChangeTab(prevPage, currentPage) {
    this.props.onChangeTab({
      i: currentPage,
      ref: this._children()[currentPage],
      from: prevPage
    });

    if (this.contentScrollDistance >= this.collapsableBarHeight) {
      this.contentView.getNode().scrollToOffset({ offset: this.collapsableBarHeight, animated: false });
    }
  },

  _handleLayout(e) {
    const { width } = e.nativeEvent.layout;

    if (Math.round(width) !== Math.round(this.state.containerWidth)) {
      this.setState({ containerWidth: width });
      this.requestAnimationFrame(() => {
        this.goToPage(this.state.currentPage);
      });
    }
  },

  _children(children = this.props.children) {
    return React.Children.map(children, child => child);
  },

  renderCollapsableBar() {
    if (!this.props.collapsibleElement) {
      return null;
    }

    return React.cloneElement(this.props.collapsibleElement, {
      key: "collapsible",
      onLayout: event => {
        this.collapsableBarHeight = event.nativeEvent.layout.height;
      }
    });
  },

  getTabBarProps(overlayTabs) {
    let tabBarProps = {
      goToPage: this.goToPage,
      tabs: this._children().map(child => child.props.tabLabel),
      activeTab: this.state.currentPage,
      scrollX: this.state.scrollX,
      scrollValue: this.state.scrollValue,
      containerWidth: this.state.containerWidth
    };

    if (this.props.tabBarBackgroundColor) {
      tabBarProps.backgroundColor = this.props.tabBarBackgroundColor;
    }
    if (this.props.tabBarActiveTextColor) {
      tabBarProps.activeTextColor = this.props.tabBarActiveTextColor;
    }
    if (this.props.tabBarInactiveTextColor) {
      tabBarProps.inactiveTextColor = this.props.tabBarInactiveTextColor;
    }
    if (this.props.tabBarTextStyle) {
      tabBarProps.textStyle = this.props.tabBarTextStyle;
    }
    if (this.props.tabBarUnderlineStyle) {
      tabBarProps.underlineStyle = this.props.tabBarUnderlineStyle;
    }
    if (overlayTabs) {
      tabBarProps.style = {
        position: "absolute",
        left: 0,
        right: 0,
        [this.props.tabBarPosition === "overlayTop" ? "top" : "bottom"]: 0
      };
    }

    return tabBarProps;
  },

  getContainerChildren() {
    const overlayTabs = this.props.tabBarPosition === 'overlayTop' || this.props.tabBarPosition === 'overlayBottom';
    const tabBarProps = this.getTabBarProps(overlayTabs);
    const children = [
      this.renderCollapsableBar()
    ];

    if (this.props.tabBarPosition === 'top') children.push(this.renderTabBar(tabBarProps));
    children.push(this.renderScrollableContent());
    if (this.props.tabBarPosition === 'bottom' || overlayTabs) children.push(this.renderTabBar(tabBarProps));

    return children;
  },

  renderFlatListItem({ item })  { return item; },

  render() {
    const children = this.getContainerChildren();
    const containerProps = {
      style: [styles.container, this.props.style, ],
      onLayout: this._handleLayout,
      ref: contentView => {this.contentView = contentView;},
      onMomentumScrollEnd: event => {this.contentScrollDistance = event.nativeEvent.contentOffset.y;}
    };

    return this.props.collapsibleElement ? (
      <AnimatedFlatList
        {...containerProps}
        stickyHeaderIndices={[1]}
        bounces={false}
        data={children}
        renderItem={this.renderFlatListItem}
        onScroll={this.props.onCollapse}
        scrollEventThrottle={1}>
      </AnimatedFlatList>
    ) : <View {...containerProps}>{ children }</View>;
  },
});

module.exports = ScrollableTabView;

const styles = StyleSheet.create({
  container: {
    flex: 1
  },
  scrollableContentAndroid: {
    flex: 1
  }
});
